/** @file		ClientFileStatus.h
 *  @brief	Keep track of files requested by or from a certain client
 *  @author	netfinity
 */
#pragma once

#include "SHAHashSet.h"
#include "SafePointers.h"
#include "NetF_Tick.h"

class CKnownFile;
class CClientFileStatusList;

/**
 *  @brief	Stores file name, comment and part status info for a file at a remote client
 */
class CClientFileStatus : public CSafeObject
{
public:
	enum EPartState
	{
		CFS_Unknown,
		CFS_Unavailable,
		CFS_Partial,
		CFS_Complete,
		CFS_Deleted
	};
	CClientFileStatus(CClientFileStatusList* list, CKnownFile* file, CUpDownClient* client);
	virtual ~CClientFileStatus();
	//
	void SetFileAICHHash(CAICHHash* hash);
	void SetPartCount(uint16 parts);
	void SetFileName(CString& name);
	void SetComment(CString& comment);
	void SetRating(uint8 rating);
	void SetCompleteSources(uint16 sources);
	CKnownFile* GetFile() const;
	CPartFile* GetPartFile() const;
	CUpDownClient* GetClient() const;
	CAICHHash* GetFileAICHHash() const;
	uint16 GetPartCount() const;
	const CString& GetFileName() const;
	const CString& GetComment() const;
	uint8 GetRating() const;
	uint16 GetCompleteSources() const;
	//
	void ClearCompleteFile();
	void ClearCompletePart(uint16 part);
	void ClearIncompletePart(uint16 part);
	void SetCompleteFile();
	void SetDeletedFile();
	void SetCompletePart(uint16 part);
	void SetSeenPart(uint16 part);
	void SetIncompletePart(uint16 part);
	void SetUploadedPart(uint16 part);
	void SetNoNeededParts() {m_bHasNeededParts = false;} //!< Sets the 'No Needed Parts' status
	void SetHasNeededParts() {m_bHasNeededParts = true;} //!< Clears the 'No Needed Parts' status
	bool IsUnknownFile() const;
	bool IsAvailableFile() const;
	bool IsCompleteFile() const;
	bool IsDeletedFile() const;
	bool IsCompletePart(uint16 part) const;
	bool IsSeenPart(uint16 part) const;
	bool IsIncompletePart(uint16 part) const;
	bool IsUploadedPart(uint16 part) const;
	int GetPartsNeededByClient(CClientFileStatus* pSrcStatus = NULL) const;
	//
	void SetUpdated();
	CTick GetLastUpdate() const;
	CTick GetLastPartUpdate() const;
	void SetDownAsked();
	CTick GetLastDownAsked() const;
	CMilliSeconds GetTimeUntilReask(const bool allowShortReaskTime = false) const;
	void SetUpRequested();
	CTick GetLastUpRequest() const;
private:
	friend					class CClientFileStatusList;
	void SetPartState(EPartState nPartState);
	CClientFileStatusList*	m_pList;						//!< Owner of this record
	CTick					m_dwLastUpdate;					//!< Time stamp for cleanup
	// File specific
	CKnownFile*				m_pFile;						//!< Pointer to file object
	uint16					m_nPartCount;					//!< Number of parts in this file
	// Client specific
	CUpDownClient*			m_pClient;						//!< Pointer to client object
	CAICHHash*  			m_pFileAICHHash;				//!< Pointer to AICH Hash for the file
	CString					m_strFilename;					//!< File name as given by client
	CString					m_strFileComment;				//!< Comment for file as given by client
	uint8					m_uFileRating;					//!< Rating for file as given by client
	CTick					m_dwLastPartUpdate;				//!< Time stamp for partstatus update
	EPartState				m_nPartState;					//!< State of part maps
	bool					m_bHasNeededParts;				//!< Client has parts we need
	uint8*					m_abitPartStatus;				//!< Client claims to have these parts
	uint8*					m_abitSeenPartStatus;			//!< Client had claimed to have these parts before
	uint8*					m_abitIncompletePartStatus;		//!< Client claims to be downloading these parts
	uint8*					m_abitUploadedPartStatus;		//!< We have uploaded these parts to the client
	uint16					m_nCompleteSourcesCount;		//!< Client claims to have seen this many complete sources
	// Upload specific
	CTick					m_dwLastUpRequest;				//!< Keep track when file was last asked by the client
	
	// Download specific
	CTick					m_dwLastDownRequest;			//!< Keep track when file was last asked from the client
	uint32      			m_nLastDownBlockOffset;			//!< Keep track of time of last requested block
	// SCT - Block maps
	// Client claims to have a set of 180k blocks available for a part
	CMap<uint16, uint16, uint64, uint64>	m_BlockMapsReceived;
	// Set of 180k blocks we have for a part and sent to client
	CMap<uint16, uint16, uint64, uint64>	m_BlockMapsSent;
};

/**
 *  @brief	Stores a list of all CClientFileStatus objects for a certain client
 */
class CClientFileStatusList
{
public:
	CClientFileStatusList(CUpDownClient* client);
	virtual ~CClientFileStatusList();
	//
	CClientFileStatus* AddFile(CKnownFile* file);
	CClientFileStatus* FindFile(const CKnownFile* file) const;
	CClientFileStatus* FindFileByHash(const uchar* fileHash) const;
	void RemoveFile(CKnownFile* file);
	void Cleanup();
	// iterators
	POSITION GetHeadPosition() const;
	CClientFileStatus* GetNext(POSITION &pos) const;
	bool IsEmpty() const;
private:
	friend							class CClientFileStatus;
	CUpDownClient*					m_pClient;					//!< Pointer to client object
	CList<CClientFileStatus*>		m_FileStatusList;
};
