/** @file		Messages.h
 *  @brief	Inter-thread communication classes
 *  @author	netfinity
 */
#pragma once
#include "stdafx.h"

class CMessageQueue;
class CMessage;

/**
 *  @brief	Message object for communication between threads
 */
class CMessage
{
public:
	CMessage(int nID, CMessageQueue* pReply = 0)
	{
		m_nID = nID; m_pReply = pReply;
	}
	virtual ~CMessage()
	{
	}
	void Reply(CMessage* pMsg);
	int		GetID()
	{
		return m_nID;
	}
private:
	CMessageQueue*	m_pReply;
	int				m_nID;
};

/**
 *  @brief	Queue for communication between threads
 */
class CMessageQueue
{
public:
	CMessageQueue()
	{
		m_bQueueIsEmpty = true;
	}
	~CMessageQueue()
	{
		PurgeMessages();
	}
	CMessage* GetMessage()
	{
		if (m_bQueueIsEmpty)
			return 0;
		CMessage*	pMsg = NULL;
		m_lockMessageList.Lock();
		if (!m_listMessages.IsEmpty())
		{
			pMsg = m_listMessages.RemoveHead();
			if (m_listMessages.IsEmpty())
				m_bQueueIsEmpty = true;
		}
		m_lockMessageList.Unlock();
		return pMsg;
	}
	void PostMessage(CMessage* pMsg)
	{
		m_lockMessageList.Lock();
		m_listMessages.AddTail(pMsg);
		if (m_bQueueIsEmpty)
			m_evtMessageReady.SetEvent();
		m_bQueueIsEmpty = false;
		m_lockMessageList.Unlock();
	}
	void PurgeMessages()
	{
		CMessage*	pMsg = NULL;
		m_lockMessageList.Lock();
		while (!m_listMessages.IsEmpty())
		{
			pMsg = m_listMessages.RemoveHead();
			delete pMsg;
		}
		m_bQueueIsEmpty = true;
		m_lockMessageList.Unlock();
	}
	void WaitMessage()
	{
		if (!m_bQueueIsEmpty)
			return;
		m_evtMessageReady.Lock();
	}
private:
	bool				m_bQueueIsEmpty;
	CEvent				m_evtMessageReady;
	CCriticalSection	m_lockMessageList;
	CList<CMessage*>	m_listMessages;
};

inline
void
CMessage::Reply(CMessage* pMsg)
{
	if (m_pReply)
		m_pReply->PostMessage(pMsg);
	else
		delete pMsg;
}

#include "emule.h"
#include "emuleDlg.h"
#include "Packets.h"
#include "EMSocket.h"
#include "ListenSocket.h"
#include "PartFile.h"

#include "NetF.h"
#include "SafePointers.h"
#include "NetF_Bytes.h"

class CEMSocket;
class CClientReqSocket;
class CUpDownClient;
class CKnownFile;

/**
 *  @brief	Message object for processing received eMule packets
 */
class CRcvPacket_Msg : public CMessage
{
public:
	const static int ID = 1;
	CRcvPacket_Msg(CEMSocket* pSocket, Packet* pPkt, uint32 nIP = 0, uint16 nPort = 0) : CMessage(ID)
	{
		m_pSocket = pSocket;
		m_pPkt = pPkt;
	}
	virtual ~CRcvPacket_Msg()
	{
		delete m_pPkt;
		CMessage::~CMessage();
	}
	CEMSocket* GetSocket()
	{
		return m_pSocket;
	}
	Packet* GetPacket()
	{
		return m_pPkt;
	}
private:
	CSafePtr<CEMSocket> 	m_pSocket;
	Packet* 				m_pPkt;
	uint32					m_nIP;
	uint16					m_nPort;
};

/**
 *  @brief	Message object for processing received eMule packets
 */
class CSockDisconnect_Msg : public CMessage
{
public:
	const static int ID = 11;
	CSockDisconnect_Msg(CEMSocket* pSocket, CString pszReason, EReason nReason) : CMessage(ID)
	{
		m_pSocket = pSocket;
		m_pszReason = pszReason;
		m_nReason = nReason;
	}
	virtual ~CSockDisconnect_Msg()
	{
		CMessage::~CMessage();
	}
	CEMSocket* GetSocket()
	{
		return m_pSocket;
	}
	CString GetReasonString()
	{
		return m_pszReason;
	}
	EReason GetReason()
	{
		return m_nReason;
	}
private:
	CSafePtr<CEMSocket> 	m_pSocket;
	CString 				m_pszReason;
	EReason 				m_nReason;
};

/**
 *  @brief	Message object for reading from disk
 */
class CIOReadReq_Msg : public CMessage
{
public:
	const static int ID = 111;
	CIOReadReq_Msg(CUpDownClient* pClient, CKnownFile *pFile, struct Requested_Block_Struct *pReqblock, byte* pBuffer, CMessageQueue* pMsgQueue) : CMessage(ID, pMsgQueue)
	{
		m_pClient = pClient;
		m_pFile = pFile;
		m_pReqblock = pReqblock;
		m_pBuffer = pBuffer;
	}
	virtual ~CIOReadReq_Msg()
	{
		CMessage::~CMessage();
	}
	CUpDownClient* GetClient()
	{
		return m_pClient;
	}
	CKnownFile* GetFile()
	{
		return m_pFile;
	}
	struct Requested_Block_Struct* GetReqblock()
	{
		return m_pReqblock;
	}
	CBytes GetOffset()
	{
		return m_pReqblock->StartOffset;
	}
	CBytes GetLength()
	{
		return m_pReqblock->EndOffset - m_pReqblock->StartOffset;
	}
	byte* GetBuffer()
	{
		return m_pBuffer;
	}
private:
	CUpDownClient*					m_pClient;
	CKnownFile*						m_pFile;
	struct Requested_Block_Struct*	m_pReqblock;
	byte*							m_pBuffer;
};

/**
 *  @brief	Message object for reading from disk
 */
class CIOReadRsp_Msg : public CMessage
{
public:
	const static int ID = 112;
	CIOReadRsp_Msg(CUpDownClient* pClient, CKnownFile *pFile, struct Requested_Block_Struct *pReqblock, byte* pBuffer) : CMessage(ID)
	{
		m_pClient = pClient;
		m_pFile = pFile;
		m_pReqblock = pReqblock;
		m_pBuffer = pBuffer;
	}
	virtual ~CIOReadRsp_Msg()
	{
		CMessage::~CMessage();
	}
	CUpDownClient* GetClient()
	{
		return m_pClient;
	}
	CKnownFile* GetFile()
	{
		return m_pFile;
	}
	struct Requested_Block_Struct* GetReqblock()
	{
		return m_pReqblock;
	}
	byte* GetBuffer()
	{
		return m_pBuffer;
	}
private:
	CUpDownClient*					m_pClient;
	CKnownFile*						m_pFile;
	struct Requested_Block_Struct*	m_pReqblock;
	byte*							m_pBuffer;
};

/**
 *  @brief	Message object for reading from disk
 */
class CIOReadErr_Msg : public CMessage
{
public:
	const static int ID = 113;
	CIOReadErr_Msg(CUpDownClient* pClient, CKnownFile *pFile, struct Requested_Block_Struct* pReqblock, byte* pBuffer, CString strCause) : CMessage(ID)
	{
		m_pClient = pClient;
		m_pFile = pFile;
		m_pReqblock = pReqblock;
		m_strCause = strCause;
		m_pBuffer = pBuffer;
	}
	virtual ~CIOReadErr_Msg()
	{
		CMessage::~CMessage();
	}
	CUpDownClient* GetClient()
	{
		return m_pClient;
	}
	CKnownFile* GetFile()
	{
		return m_pFile;
	}
	struct Requested_Block_Struct* GetReqblock()
	{
		return m_pReqblock;
	}
	CString GetCause()
	{
		return m_strCause;
	}
	byte* GetBuffer()
	{
		return m_pBuffer;
	}
private:
	CUpDownClient*					m_pClient;
	CKnownFile*						m_pFile;
	struct Requested_Block_Struct*	m_pReqblock;
	CString							m_strCause;
	byte*							m_pBuffer;
};